#ifdef MAP_REGEX

#include "sendmail.h"
#include <sys/types.h>
#include "regex.h"

#define DEFAULT_DELIM	"@"

#define ERRBUF_SIZE	80
#define RETBUF_SIZE	256
#define MAX_MATCH	32

#define MF_REGEX_NOT	0x00040000 /* move to sendmail.h */

#define xnalloc(s)	memset(xalloc(s), 0, s);

struct regex_map {
regex_t pattern_buf; /* xalloc it */
long regex_flags; /* move to map->map_mflags */
int regex_substrings; /* move to type MAP */
char *delim; /* move to type MAP */
};

int regex_map_init(map, ap)
	MAP *map;
	char *ap;
{
	int regerr;
	struct regex_map *map_p;
	register char *p;
	int pflags;

	if(tTd(72, 1))
		printf("regex_map_init: mapname '%s', args '%s'\n", 
				map->map_mname, ap);

	pflags = REG_ICASE | REG_EXTENDED | REG_NOSUB;

	p = ap;

	map_p = xnalloc(sizeof(struct regex_map));

	for (;;)
        {
		while (isascii(*p) && isspace(*p))
			p++;
		if (*p != '-')
			break;
		switch (*++p)
		{
			case 'n': /* not */
				map_p->regex_flags = MF_REGEX_NOT;
				break;

			case 'f': /* case sensitive */
				map->map_mflags |= MF_NOFOLDCASE;
				pflags &= ~REG_ICASE;
				break;

			case 'b': /* basic regular expressions */
				pflags &= ~REG_EXTENDED;
				break;

			case 's': /* substring match () syntax */
				pflags &= ~REG_NOSUB;
				break;

			case 'd': /* delimiter */
				map_p->delim = ++p;
				break;

			case 'a': /* map append */
				map->map_app = ++p;
				break;
		}
                while (*p != '\0' && !(isascii(*p) && isspace(*p)))
                        p++;
                if (*p != '\0')
                        *p++ = '\0';
	}
	if(tTd(72, 1))
		printf("regex_map_init: compile '%s' %x\n", p, pflags);

	if((regerr = regcomp(&(map_p->pattern_buf), p, pflags)) != 0)
	{ /* Errorhandling */
		char errbuf[ERRBUF_SIZE];
		regerror(regerr, &(map_p->pattern_buf), errbuf, ERRBUF_SIZE);
		syserr("pattern-compile-error: %s\n", errbuf);
		free(map_p);
		return FALSE;	
	}

	if (map->map_app != NULL)
		map->map_app = newstr(map->map_app);
	if (map_p->delim != NULL)
		map_p->delim = newstr(map_p->delim);
	else
		map_p->delim = DEFAULT_DELIM;

	if((pflags & REG_NOSUB) == 0)
	{ /* substring matching */
		char *cp;
		char lastc = '\0';
		for(cp = p; *cp; cp++)
		{
			if(*cp == '(' && lastc != '\\')
				map_p->regex_substrings++;

			if(*cp == '\\' && lastc == '\\')
				lastc = '\0';
			else
				lastc = *cp;
		}
		if(tTd(72, 1))
			printf("regex_map_init: substrings %d\n", 
					map_p->regex_substrings);
	}

	map->map_file = (char *)map_p; /* dirty hack */

	return TRUE;
}

char *regex_map_lookup(map, name, av, statp)
	MAP *map;
	char *name;
	char **av;
	int *statp;
{
	int regret;
	struct regex_map *map_p;
	regmatch_t pmatch[MAX_MATCH];

	if(tTd(72, 1))
		printf("regex_map_lookup: key '%s'\n", name);

	map_p = (struct regex_map *)(map->map_file);
	regret = regexec(&(map_p->pattern_buf),
			 name, MAX_MATCH, pmatch, 0);
	
	if((map_p->regex_flags & MF_REGEX_NOT) != 0)
	{ /* option -n */
		if(regret == REG_NOMATCH)
			return map_rewrite(map, "", 0, av);
		else
			return NULL;
	}
	if(regret == REG_NOMATCH)
		return NULL;

	if(map_p->regex_substrings > 0)
	{ /* option -s */
		static char retbuf[RETBUF_SIZE];
		register char *dp, *sp;
		char *endp;
		int i;

		dp = retbuf;

		for(i = 1; i < map_p->regex_substrings + 1; i++)
		{
			if(i > 1)
				for(sp = map_p->delim; *sp; sp++)
					*dp++ = *sp;

			if(pmatch[i].rm_so < 0 || pmatch[i].rm_eo < 0)	
				continue;
			
			for(sp = name + pmatch[i].rm_so, 
			    endp = name + pmatch[i].rm_eo;
			    endp - sp > 0; 
			    sp++)
				*dp++ = *sp;
		}

		*dp = '\0';

		return map_rewrite(map, retbuf, strlen(retbuf), av);
	}
	return map_rewrite(map, "", 0, av);
}
#endif /* MAP_REGEX */
